;; 
;;    $Id: alarm-clock.asm,v 1.6 2003/11/12 11:13:17 james Exp $
;; 
;;    alarm-clock.asm, a PIC 16F877-04P with LCD as an alarm clock.
;;    Copyright (C) 2002  James Cameron (quozl@us.netrek.org)
;;
;;    This program is free software; you can redistribute it and/or modify
;;    it under the terms of the GNU General Public License as published by
;;    the Free Software Foundation; either version 2 of the License, or
;;    (at your option) any later version.
;;
;;    This program is distributed in the hope that it will be useful,
;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;    GNU General Public License for more details.
;;
;;    You should have received a copy of the GNU General Public License
;;    along with this program; if not, write to the Free Software
;;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;;

;;
;;    Hardware assumptions
;;
;;    LCD panel connected via Dontronics DT106 PCB
;;    4MHz crystal oscillator
;;
	
	processor	16f877
	list		f=inhx8m
	include		"p16f877.inc"

	__config	_lvp_off & _wdt_off & _hs_osc

	;; symbol naming conventions
	;;	p_ port address of
	;;	t_ tris address of
	;;	m_ mask bits of
	;;	b_ bit number
	;;	c_ command mask
	
	;; file register allocations
	
base		equ	0x20		; first free file register address
stack		equ	0x7f+1		; end address of stack

	;; port bit allocations (a Dontronics DT106 PCB)

p_lcdc		equ	portd		; lcd control port address
t_lcdc		equ	trisd		; lcd control tris address
b_lcd_rs	equ	1		; lcd register select (pin 4)
b_lcd_rw	equ	2		; lcd read/write (pin 5)
b_lcd_e		equ	3		; lcd enable (pin 6)
m_lcdc		equ	b'00001110'	; lcd control mask bits

p_lcdd		equ	portd		; lcd data port address
t_lcdd		equ	trisd		; lcd data tris address
b_lcd_db4	equ	4
b_lcd_db5	equ	5
b_lcd_db6	equ	6
b_lcd_db7	equ	7
m_lcdd		equ	b'11110000'	; lcd data mask bits

	cblock	base
		r_ephemeral		; use in lowest level functions only
		r_temporary		; use in lowest level functions only
		r_trisl			; mirror of lcd port tris
		flag:2			; execution flags
	endc

#define flag_1hz flag,0	; set by isr each second
#define flag_kb flag,1		; set by isr on keyboard input
#define flag_alarm flag,2	; set by isr on alarm trigger
	
#define flag_next flag,3	; whether to wake the sleeper
#define flag_dst flag,4		; dst time shift in effect
#define flag_hash flag,5	; hash key pressed, waiting for opcode
#define flag_tick flag,6	; whether to tick tock or not
#define flag_beeps flag,7	; keyboard beep mode

#define flag_temp flag+1,0	; show temperature on display
#define flag_holiday flag+1,1	; holiday mode
#define flag_ring flag+1,2	; alarm ringing

	include	"stack.inc"		; stack macros

align	macro
	org	(($>>2)<<2)+4		; align to four byte boundary
	endm
	
	;; reset vector
	org	0x0
	goto	main			; go to main program

	;; interrupt vector
	org	0x4
	goto	isr

table	macro
	popw				; grab offset
	addwf	pcl,f			; vector to appropriate retlw
	endm

table_end	macro
	retlw	0			; terminating zero for caller
	endm
	
t_revision
	table
	dt	"$Revision: 1.6 $"
	;        ----------------
	table_end

t_welcome
	table
	dt	"Diurnal Waker"
	;        ----------------
	table_end

t_alarm
	table
	dt	"Alarm"
	table_end

t_silence
	table
	dt	"Silence"
	table_end

t_wakeup
	table
	dt	"Wakeup boss!"
	table_end

t_keyboard
	table
	dt	"Keyboard, "
	table_end

t_set_time
	table
	dt	"Set time, "
	table_end

t_daylight
	table
	dt	"Daylight, "
	table_end

t_spring
	table
	dt	"Spring"
	table_end

t_autumn
	table
	dt	"Autumn"
	table_end

t_temp
	table
	dt	"Temperature"
	table_end

t_tick
	table
	dt	"Tick tock"
	table_end

t_beeps
	table
	dt	"Key beeps"
	table_end

t_holiday
	table
	dt	"Holiday"
	table_end

t_on
	table
	dt	" on"
	table_end

t_off
	table
	dt	" off"
	table_end

	;; interrupt service routine
isr

	cblock
		saved_w			; saved w register
		saved_status		; saved status register (swapped)
		saved_fsr		; saved fsr register
	endc

	movwf	saved_w			; save w register
	swapf	status,w		; save status register
	movwf	saved_status

	bcf	status,rp0
	bcf	status,rp1
	
	btfss	intcon,t0if		; test timer zero interrupt flag
	goto	isr_skip_t0if

isr_clock
	bcf	intcon,t0if		; clear timer zero interrupt flag

	cblock
		u0			; microsecond counter low
		u1			; microsecond counter middle
		u2			; microsecond counter high
	endc

	;; add 65536us to the microsecond counter
	incf	u2,f
	
	cblock
		t0			; temporary copy of u0-u2
		t1
		t2
	endc

	;; save a copy of the counter
	mov24	u0,t0
	
	;; subtract 1s from copy of counter
	movlw	d'1000000'&0xff
	subwf	t0,f	
	btfss	status,c
	decf	t1,f	
	
	movlw	d'1000000'>>d'8'&0xff
	subwf	t1,f	
	btfss	status,c
	decf	t2,f	
	
	movlw	d'1000000'>>d'16'&0xff
	subwf	t2,f	
	btfss	status,c
	goto	isr_clock_end		; underflow, ignore the subtraction

	;; did not underflow, save the result, occurs once per second
	mov24	t0,u0

	cblock
		ss			; seconds
		mm			; minutes
		hh			; hours
		dl			; days
		dh
	endc

	;; save fsr
	movf	fsr,w
	movwf	saved_fsr
	
	;; point to counter
	movlw	ss
	movwf	fsr

	incf	indf,f			; increment second counter
	
	movlw	d'60'
	subwf	indf,w
	bnz	isr_kick_end		; hasn't reached a minute

	clrf	indf			; clear the second counter
	incf	fsr,f			; point to minute counter
	incf	indf,f			; increment minute counter
	
	movlw	d'60'
	subwf	indf,w
	bnz	isr_kick_end		; hasn't reached an hour

	clrf	indf			; clear the minute counter
	incf	fsr,f			; point to hour counter
	incf	indf,f			; increment hour counter

	movlw	d'24'
	subwf	indf,w
	bnz	isr_kick_end		; hasn't reached an hour

	clrf	indf			; clear the hour counter
	incf	fsr,f			; point to day counter
	incf	indf,f			; increment day counter
	
	bnz	isr_kick_end
	incf	fsr,f
	incf	indf,f

isr_kick_end
	movf	ss,w		; don't check if seconds non-zero
	bnz	isr_next_end
	movf	wm,w		; check minute, it must match
	subwf	mm,w
	bnz	isr_next_end
	movf	wh,w		; check hour, it must match
	subwf	hh,w
	bnz	isr_next_end
	btfsc	flag_next	; if alarm enabled
	bsf	flag_alarm	; trigger the alarm
	btfss	flag_holiday	; if not a holiday
	bsf	flag_next	; enable alarm for next day
	
isr_next_end	
	;; restore fsr
	movf	saved_fsr,w
	movwf	fsr

	;; wake mainline
	bsf	flag_1hz
	
isr_clock_end
	
isr_skip_t0if
	btfss	intcon,rbif		; if port change not detected
	goto	isr_skip_rbif		; jump over the handler
	
	call	kb_isr			; call keyboard interrupt routine
	bcf	intcon,rbif		; clear interrupt flag
	
isr_skip_rbif
	swapf	saved_status,w		; restore registers saved
	movwf	status
	swapf	saved_w,f
	swapf	saved_w,w
	retfie

	
	include	"table.asm"		; table handler
	include	"lcd.asm"		; lcd display functions
	include	"bcd.asm"		; binary to decimal conversion
	include	"hex.asm"		; binary to hexadecimal conversion
	include	"stack.asm"		; stack functions
	;; include	"ad.asm"		; analog to digital conversion
	include	"sb.asm"		; sixteen bit math functions
	;; include	"ee.asm"		; eeprom functions
	include	"delay.asm"		; delay functions
	include	"pwm.asm"		; pwm sound output functions
	include	"kb.asm"		; keyboard functions
	include	"ts.asm"		; temperature sensor
	include	"ow.asm"		; one wire bus
		
	;; 0123456789012345
	;; $Revision: 1.6 $
	;; * hh:mm:ss tt.tc
	;; Message .......#
pos_welcome	equ	lcd_row_0
pos_revision	equ	lcd_row_1
pos_data	equ	lcd_row_0
pos_message	equ	lcd_row_1
pos_days	equ	pos_data+2
pos_next	equ	pos_data+0
pos_temp	equ	pos_data+d'11'
pos_hash	equ	pos_message+d'15'
	
	
initialise
	;; clear stack
	movlw	stack
	movwf	fsr
	
	;; clear output port
	clrf	porta
	clrf	portb
	clrf	portc
	clrf	portd
	clrf	porte

	;; set tristate latches
	movlw	~(m_lcdd|m_lcdc)	; set lcd data and control to output
	movwf	r_trisl
	
	bsf	status,rp0
	movwf	trisd
	
	movlw	b'11111111'		; inputs so we can use analog
	movwf	trisa
	
	movlw	b'00000000'		; rest outputs
	movwf	trisb
	movwf	trisc
	movwf	trise

	bcf	status,rp0

	;; clear counters
	clr24	u0
	clrf	ss
	clrf	mm
	clrf	hh
	clr16	dl

	clrf	flag			; clear synchronisation flags
	clrf	flag+1			; clear synchronisation flags

	;; clear wake time
	movlw	d'8'			; 08:00
	movwf	wh
	clrf	wm

	;; initialise option register
	bsf	status,rp0
	bcf	option_reg,t0cs		; set timer to use oscillator/4
	bcf	option_reg,psa		; set prescaler to TMR0
	bsf	option_reg,ps2		; set prescaler to 1:256
	bsf	option_reg,ps1
	bsf	option_reg,ps0
	bcf	status,rp0

	;; lastly, enable interrupts
	bsf	intcon,t0ie		; enable timer interrupt
	bsf	intcon,gie		; enable all interrupts

	bsf	portd,0
	;; initialise lcd
	call	lcd_initialise

	;; initialise keyboard
	call	kb_initialise

	;; initialise input buffer
	goto	ib_initialise
	;; return

	cblock
		msg_count
	endc

	;; begin a message, reset timer, clear the line.
msg_begin
	movlw	d'5'		; number of seconds to show a message
	movwf	msg_count
	
	;; clear the message line and leave cursor at start of line
msg_clear			; ( -- )
	call	msg_home
	call	lcd_put_spaces16

	;; position cursor to message line
msg_home			; ( -- )
	pushl	pos_message
	call	lcd_move
	return

	;; clear the message line and show a message
msg_put	macro	table
	call	msg_clear
	pers	lcd_put,table
	endm

	;; add a message to the current position
msg_add	macro	table
	pers	lcd_put,table
	endm

	;; age the message shown, clear once it is old enough
msg_aging
	movf	msg_count,f
	btfsc	status,z
	return

	decfsz	msg_count,f
	return

	goto	msg_clear

show_welcome
	pushl	pos_welcome
	call	lcd_move
	msg_add	t_welcome
	
	pushl	pos_revision
	call	lcd_move
	msg_add	t_revision
	return

	
isr_interlock_on	macro
	bcf	intcon,gie		; disable interrupts
	btfsc	intcon,gie		; test we suceeded [paranoia]
	goto	$-2			; loop until it worked
	endm

isr_interlock_off macro
	bsf	intcon,gie		; enable interrupts
	endm
	
	;; get time, interlocked
	;; ( -- ss mm hh )
get_time
	isr_interlock_on
	pushf	ss
	pushf	mm
	pushf	hh
	isr_interlock_off
	return

	;; set time, interlocked
	;; ( ss mm hh -- )
set_time
	isr_interlock_on
	popf	hh
	popf	mm
	popf	ss
	isr_interlock_off
	return
	
	
	;; display a time at the current cursor position
	;; ( mm hh -- mm hh )
lcd_put_time_hh_mm_no_drop
	over
	over
	
	;; display a time at the current cursor position
	;; ( mm hh -- )
lcd_put_time_hh_mm
	pushl	d'0'			; hh
	call	sb_bcd
	drop
	drop
	call	lcd_put_hex
	
	pushl	':'
	call	lcd_put

	pushl	d'0'			; mm
	call	sb_bcd
	drop
	drop
	goto	lcd_put_hex

	;; update the display
dsp_all
	call	dsp_next
	call	dsp_time
	call	dsp_temp
	goto	msg_aging
	
dsp_next
	pushl	pos_next
	call	lcd_move
	pushl	'*'		; emit the star if alarm is set
	movlw	' '
	btfss	flag_next
	movwf	indf
	goto	lcd_put

dsp_time
	pushl	pos_days
	call	lcd_move
	call	get_time	; ( -- ss mm hh )
	goto	lcd_put_time	; ( ss mm hh -- )
	
dsp_temp
	;; either erase field or display temperature
	pushl	pos_temp
	call	lcd_move
	
	btfsc	flag_temp
	goto	ts_get

	pushl	d'5'
	goto	lcd_put_spaces
	
	;; display a time at the current cursor position
	;; ( ss mm hh -- )
lcd_put_time
	call	lcd_put_time_hh_mm
	
	pushl	':'
	call	lcd_put

	pushl	d'0'			; ss
	call	sb_bcd
	drop
	drop
	goto	lcd_put_hex


ib_size	equ	4
	cblock
		ib:ib_size
		ib_digits
	endc

ib_initialise			; ( -- )
	clrf	ib		; reset the input buffer to be empty
	clrf	ib+1
	clrf	ib+2
	clrf	ib+3
	clrf	ib_digits
	return

ib_clear			; ( -- )
	call	ib_initialise	; wipe the buffer contents
	goto	msg_clear
		
ib_put				; ( key -- )
	incf	ib_digits,f	; count the digits entered
	nop
	nop
	;; ??? limit counter to ib_size 
	movf	ib+2,w
	movwf	ib+3
	movf	ib+1,w
	movwf	ib+2
	movf	ib,w
	movwf	ib+1

	movlw	'0'		; adjust from ascii to digit
	subwf	indf,f
	popw			; place the new digit into position
	movwf	ib
	
ib_show				; ( -- )
	msg_put	t_keyboard
	movlw	d'65'		; number of seconds to keyboard input
	movwf	msg_count
	pushf	ib+3
	call	lcd_put_digit
	pushf	ib+2
	call	lcd_put_digit
	pushl	':'
	call	lcd_put
	pushf	ib+1
	call	lcd_put_digit
	pushf	ib
	goto	lcd_put_digit

ib_get				; ( -- 0 | h h m m n )
	pushf	ib_digits
	movf	indf,f		; set zero flag from stack top
	skpnz
	return
	
ib_get_data
	pop
	pushf	ib+3
	pushf	ib+2
	pushf	ib+1
	pushf	ib
	pushf	ib_digits
	;; ??? caller is expected to call ib_clear
	return
		
	cblock
		wm		; wake time in minutes
		wh		; wake time in hours
	endc

	;; cancel alarm, set wake time or toggle next day wake flag
op_star				; ( -- )
	btfss	flag_ring	; if not ringing alarm
	goto	op_star_ok
	
	bcf	flag_ring	; turn off alarm
	call	ib_clear	; and wipe input buffer
	call	msg_begin
	msg_put	t_silence
	return
	
op_star_ok
	call	ib_get		; get input buffer (if any)
	call	ib_clear	; and wipe it
	call	msg_begin
	msg_put	t_alarm
	movf	indf,f		; set zero flag according to count
	bz	op_star_toggle
	pop

pop_two_digits_to_f	macro	wm
	popf	wm
	bcf	status,c
	rlf	indf,f		; x 2
	peekw			; grab
	rlf	indf,f		; x 4
	rlf	indf,f		; x 8
	addwf	indf,w		; x 8 + x 2 = x 10
	addwf	wm,f		; include in minutes
	pop
	endm

	;; set wake time
	pop_two_digits_to_f wm
	pop_two_digits_to_f wh
	goto	op_star_set
	
	;; toggle next day wake flag
op_star_toggle
	pop
	btfss	flag_next	; test flag
	goto	op_star_set

op_star_clear
	bcf	flag_next
	msg_add	t_off
	call	dsp_next
	call	pwm_no
	return
	
op_star_set
	bsf	flag_next
	msg_add	t_on

	pushl	','
	call	lcd_put
	call	lcd_put_space
	pushf	wm
	pushf	wh
	call	lcd_put_time_hh_mm
	call	dsp_next
	call	pwm_yes
	return
	

op_set_time			; ( -- )
	call	ib_get		; get the input buffer
	call	ib_clear	; and wipe it
	call	msg_begin
	movf	indf,f		; set zero flag according to count
	bz	op_set_time_dst	; if no arguments, do dst toggle
	msg_put	t_set_time

	decf	indf,f		; set zero flag if only one digit entered
	bz	op_set_time_hh
	decf	indf,f		; set zero flag if only two digits entered
	bz	op_set_time_hh

op_set_time_mm
	pop
	pop_two_digits_to_f r_ephemeral
	pop_two_digits_to_f r_temporary
	pushl	d'0'		; ss
	pushf	r_ephemeral	; mm
	pushf	r_temporary	; hh
	call	lcd_put_time_hh_mm_no_drop
	goto	op_set_time_do
	
op_set_time_hh
	pop
	call	get_time	; ( ss mm hh -- )
	drop			; lose the hours
	pop_two_digits_to_f r_ephemeral
	call	lcd_put_time_hh_mm_no_drop
	goto	op_set_time_do
	
op_set_time_dst
	pop
	msg_put	t_daylight
	btfss	flag_dst
	goto	op_set_time_dst_set
	
	;; we are in dst time now, switching out
	bcf	flag_dst
	call	get_time
	decf	indf,f
	;; ??? improper handling of underflow
	msg_add	t_autumn
	goto	op_set_time_do
	
op_set_time_dst_set
	bsf	flag_dst
	call	get_time
	incf	indf,f
	;; ??? improper handling of overflow
	msg_add	t_spring

op_set_time_do
	call	set_time	; ( ss mm hh -- )
	call	dsp_time
	call	pwm_yes
	return

	;; toggle temperature display
op_temp
	call	ib_clear	; clear any input buffer
	call	msg_begin
	msg_put	t_temp

	btfss	flag_temp
	goto	op_temp_set

op_temp_clear
	bcf	flag_temp
	msg_add	t_off
	call	dsp_temp
	call	pwm_no
	return

op_temp_set
	bsf	flag_temp
	msg_add	t_on
	call	dsp_temp
	call	pwm_yes
	return
	
	;; toggle tick tock 
op_tick
	call	ib_clear	; clear any input buffer
	call	msg_begin
	msg_put	t_tick

	btfss	flag_tick
	goto	op_tick_set

op_tick_clear
	bcf	flag_tick
	msg_add	t_off
	call	pwm_no
	return

op_tick_set
	bsf	flag_tick
	msg_add	t_on
	call	pwm_yes
	return
	
	;; toggle keyboard beeps
op_beeps
	call	ib_clear	; clear any input buffer
	call	msg_begin
	msg_put	t_beeps

	btfss	flag_beeps
	goto	op_beeps_set

op_beeps_clear
	bcf	flag_beeps
	msg_add	t_off
	call	pwm_no
	return

op_beeps_set
	bsf	flag_beeps
	msg_add	t_on
	call	pwm_yes
	return

	;; test, trigger alarm
op_alarm
	bsf	flag_alarm
	goto	ib_clear
	
	;; toggle holiday mode
op_holiday
	call	ib_clear	; clear any input buffer
	call	msg_begin
	msg_put	t_holiday

	btfss	flag_holiday
	goto	op_holiday_set

op_holiday_clear
	bcf	flag_holiday
	msg_add	t_off
	call	pwm_no
	return

op_holiday_set
	bsf	flag_holiday
	msg_add	t_on
	call	pwm_yes
	return
	
;;;
;;; main program
;;; loop is woken by isr to process events
;;; 

main
	call	initialise
	call	show_welcome
	btfss	flag_1hz	; wait a second
	goto	$-1
	call	lcd_clear
	call	handle_1hz
	
	;; spin until an event occurs requiring action
loop
	btfsc	flag_1hz
	call	handle_1hz
	
	btfsc	flag_alarm
	call	handle_alarm
	
	btfsc	flag_kb
	call	handle_kb
	
	goto	loop

	
	;; handle 1hz update events
handle_1hz
	bcf	flag_1hz
	call	dsp_all

	btfsc	flag_ring	; if alarm ringing, go there, don't tick
	goto	handle_ring
	
	btfss	flag_tick	; tick tock
	return
	pushl	pwm_2000
	call	pwm_tick
	return

	
	;; handle keyboard events
handle_kb
	btfss	flag_beeps
	goto	handle_kb_get
	pushl	pwm_0440
	call	pwm_brief
	
handle_kb_get
	call	kb_get		; ( -- key ) [resets flag_kb]

;	set time (hhmm#0)
;	set hour (hh#0)
;	toggle dst (#0)
;	set wake (hhmm*)
;	toggle wake (*)
	
;	operation (digits # code)

	peekw
	sublw	'*'
	bnz	not_star

	call	op_star
	
	bcf	flag_hash	; clear any pending hash code
	goto	end_kb_pop

not_star
	peekw
	sublw	'#'
	bnz	not_hash

	pushl	pos_hash	; put # on screen to acknowledge key
	call	lcd_move
	pushl	'#'
	call	lcd_put
	
	bsf	flag_hash	; next digit is opcode
	goto	end_kb_pop
		
not_hash
	btfsc	flag_hash
	goto	not_default
	
	;; accept digit into buffer
	call	ib_put
	goto	end_kb

not_default
	bcf	flag_hash

op	macro	key,action
	peekw
	sublw	key
	btfsc	status,z
	call	action
	endm
	
	op	'0',op_set_time
	op	'1',op_holiday
	op	'2',op_tick
	op	'3',op_beeps
	op	'4',op_temp
;	op	'5',ib_clear
;	op	'6',ib_clear
;	op	'7',ib_clear
;	op	'8',ib_clear
	op	'9',op_alarm
	
end_kb_pop
	pop			; ( key -- )
end_kb
	
skip_kb
	return


	;; handle alarm events
handle_alarm
	clrf	ringer
	bcf	flag_alarm
	bsf	flag_ring
	
	call	msg_begin	; show wakeup message for two minutes
	msg_put	t_wakeup
	movlw	d'120'
	movwf	msg_count

	return

handle_ring
	
	cblock
		ringer
	endc

	incfsz	ringer,f
	goto	ring_build	; not yet 256 seconds

	bcf	flag_ring	; give up
	pushl	pwm_1000
	goto	pwm_long

ring_build
	btfsc	ringer,6	; if we've done about a minute, 
	goto	ring_long

ring_tick
	decf	ringer,w
	andlw	b'111'
	skpz
	return

	pushl	pwm_0880
	goto	pwm_brief

ring_long
last
	movf	ringer,w
	andlw	b'11'
	skpz
	return
		
	pushl	pwm_0880
	call	pwm_long
	pushl	pwm_0440
	goto	pwm_brief

	if	last > 0x7ff
	messg	"error:	code runs over page (>0x73ff)"
	endif
	
	end
