;;;
;;; ow_, routines for one wire bus communication
;;;

movwt	macro			; move w to tris
        tris	t_ts
	endm

;;;
;;; ow_command_*, one wire bus command bytes, from ds-1820 data sheet
;;; 
ow_command_skip		equ	0xcc	; skip rom match sequence
ow_command_convert	equ	0x44	; commence temperature conversion
ow_command_read		equ	0xbe	; read the data scratchpad

;;;
;;; ow_high, one wire high, allow output line to float to pull-up
;;;
;;; call stack:	none
;;; data stack:	none
;;; 
ow_high
	movf	r_which,w	; fetch the selection mask
	iorwf	p_ts,f		; set that bit in port latch
				; (a momentary 2uS fast pull-up)
	iorwf	r_trism,w	; set that bit in tris mask
	movwf	r_trism		; save tris mask
	movwt			; set tris
        retlw	0

;;;
;;; ow_low, one wire low, drive output line to ground
;;;
;;; call stack:	none
;;; data stack:	none
;;; 
ow_low
	comf	r_which,w	; fetch inverted selection bits
	andwf	p_ts,f		; reset selected bits of port
	andwf	r_trism,f	; reset selected bits of tris mirror
	movf	r_trism,w	; fetch tris mirror
	movwt			; set tris
        retlw	0

;;;
;;; ow_reset, one wire reset, signal the slave that we want a bus reset
;;;
;;; call stack:	calls others that do not call
;;; data stack:	( -- slave-present-flag ) [total 1]
;;; 
ow_reset
	;; allow the bus to go high
	call	ow_high
	pushl	d'9'		; 60us delay
	call	us10
	
	;; now check to see if the bus is sick (no pull-up)
	movf	p_ts,w		; read port bits once only
	andwf	r_which,w	; keep bits we want
	bz	ow_reset_fail	; bus is being held down, fail

	;; bus is healthy, now issue a reset
	call	ow_low
	pushl	d'50'		; 500us delay (480 to 960us)
	call	us10
	call	ow_high
	pushl	d'9'		; 60us delay
	call	us10
	movf	p_ts,w		; read port bits once only
	andwf	r_which,w	; keep bits we want
	bz	ow_reset_ok
	pushl	d'39'		; total 480us delay
	call	us10
	
ow_reset_fail
	pushl	0		; return failure status
	retlw	0

ow_reset_ok
	pushl	d'39'		; total 480us delay
	call	us10
	pushl	1		; return success status
	retlw	0

;;;
;;; ow_rx, one wire receive, receive a byte from the bus
;;;
;;; call stack:	calls others that do not call
;;; data stack:	( -- received-byte ) [total 1]
;;; 
ow_rx
	pushl	d'0'		; received data
	pushl	d'8'		; count of bits to read in

ow_rx_loop
	call	ow_low		; generate short low for slave
	nop
	call	ow_high
	nop			; delay for 6 microseconds
	nop
	nop
	nop
	nop
        nop
	movf	p_ts,w		; read port bits once only
	andwf	r_which,w	; keep bits we want
	btfsc	status,z	; test if it is zero
	bcf	status,c	; it is, so clear carry
	btfss	status,z	; test if it is one
	bsf	status,c	; it is, so set carry

	incf	fsr,f		; rotate the bit into our copy
	rrf	indf,f	
	decf	fsr,f	
	
	pushl	d'6'		; and wait for sixty microseconds
	call	us10
	decfsz	indf,f		; decrement count of bits to read
	goto	ow_rx_loop	; loop until end of byte
	incf	fsr,f		; drop count of bits
	retlw	0
	
;;;
;;; ow_tx, one wire transmit, output a byte on the bus
;;;
;;; call stack:	calls others that do not call
;;; data stack:	( transmit-byte -- ) [total 3]
;;; 
ow_tx
	pushl	d'8'		; initialise bits in byte counter

ow_tx_loop
	incf	fsr,f		; rotate next bit to carry flag
	rrf	indf,f	
	decf	fsr,f	
        btfsc	status,c	; test the bit
	goto	ow_tx_one
	
ow_tx_zero			; transmit a zero
	call	ow_low		; by generating a long low
	pushl	d'6'		; for sixty microseconds
	call	us10
	call	ow_high		; then return to high
	goto	ow_tx_skip	; and continue to next bit
	
ow_tx_one			; transmit a one
	call	ow_low		; by generating a short low
	call	ow_high		; followed by a long high
	pushl	d'6'		; for sixty microseconds
	call	us10

ow_tx_skip
	decfsz	indf,f		; decrement bits in byte counter
        goto	ow_tx_loop	; loop until counter zero
	incf	fsr,f		; drop bits in byte counter
	incf	fsr,f		; drop transmit byte
	retlw	0		; return to caller

